package main

import (
	"context"
	"encoding/hex"
	"errors"
	"flag"
	"fmt"
	"strconv"
	"strings"
	"time"

	"sub_demo/proto/subscribe"
	"sub_demo/sasl_sarama"

	"github.com/Shopify/sarama"
	mapset "github.com/deckarep/golang-set"
	"github.com/golang/protobuf/proto"
)

const (
	PRINT_NONE      = 0
	PRINT_DMLOPT    = 1
	PRINT_KAFKAMETA = 2
)

var (
	partitionConsumers = make(map[int32]*tdsqlPartitionMsgConsumer)
	brokerList         = flag.String("brokers", "", "REQUIRED: The comma separated list of brokers in the Kafka cluster")
	topic              = flag.String("topic", "", "REQUIRED: the topic to consume")
	group              = flag.String("group", "", "REQUIRED: the consumer group")
	user               = flag.String("user", "", "the user name")
	password           = flag.String("password", "", "the password")
	trans2sql          = flag.Bool("trans2sql", false, "trans message to sql or not")
	verbose            = flag.Int("verbose", PRINT_DMLOPT, "whether to turn on debugging logging")
)

var numberTypes = mapset.NewSet(subscribe.DataType_INT8, subscribe.DataType_UINT8,
	subscribe.DataType_INT16, subscribe.DataType_UINT16,
	subscribe.DataType_INT32, subscribe.DataType_UINT32,
	subscribe.DataType_INT64, subscribe.DataType_UINT64,
	subscribe.DataType_FLOAT32, subscribe.DataType_FLOAT64,
	subscribe.DataType_DECIMAL)

func main() {
	flag.Parse()
	config := sarama.NewConfig()
	config.Version = sarama.V0_11_0_0
	config.Consumer.Return.Errors = true
	config.Consumer.Offsets.AutoCommit.Enable = true //not to auto-commit updated offsets back to the broker.
	config.Consumer.Offsets.Initial = sarama.OffsetNewest
	config.Consumer.Group.Rebalance.Strategy = sarama.BalanceStrategyRange
	//消费者未上报心跳超过此时间则认为宕机，触发 rebalance
	config.Consumer.Group.Rebalance.Timeout = 60 * time.Second
	config.ChannelBufferSize = 1000

	if *user != "" {
		config.Net.SASL.Enable = true
		config.Net.SASL.User = *user
		config.Net.SASL.Password = *password
		config.Net.SASL.Handshake = true
		config.Net.SASL.SCRAMClientGeneratorFunc = func() sarama.SCRAMClient { return &sasl_sarama.XDGSCRAMClient{HashGeneratorFcn: sasl_sarama.SHA512} }
		config.Net.SASL.Mechanism = sarama.SASLTypeSCRAMSHA512
	}

	// Start with a client
	client, err := sarama.NewClient(strings.Split(*brokerList, ","), config)
	if err != nil {
		panic(err)
	}
	defer func() { _ = client.Close() }()

	// Start a new consumer group
	group, err := sarama.NewConsumerGroupFromClient(*group, client)
	if err != nil {
		panic(err)
	}
	defer func() { _ = group.Close() }()

	// Track errors
	go func() {
		for err := range group.Errors() {
			fmt.Println("ERROR", err)
		}
	}()

	// Iterate over consumer sessions.
	ctx := context.Background()
	for {
		handler := demoTdsqlConsumerGroupHandler{
			verbose: *verbose,
		}

		err := group.Consume(ctx, []string{*topic}, handler)
		if err != nil {
			panic(err)
		}

	}
}

type demoTdsqlConsumerGroupHandler struct {
	verbose int //open debug
}

func (h demoTdsqlConsumerGroupHandler) Setup(sess sarama.ConsumerGroupSession) error {
	fmt.Println("setup:", sess.Claims()) //获取到所有的 topic 和 partition

	setup := make(map[int32]bool)
	for _, part := range sess.Claims()[*topic] {
		if _, ok := partitionConsumers[part]; !ok {
			partitionConsumers[part] = &tdsqlPartitionMsgConsumer{
				shardLastSeqMap: map[string]uint64{},
				shardMsgMap:     map[string]*subscribe.Envelope{},
				verbose:         h.verbose,
			}
		}

		setup[part] = true
	}

	var invalid []int32
	for p := range partitionConsumers {
		if _, ok := setup[p]; !ok {
			invalid = append(invalid, p)
		}
	}

	for _, p := range invalid {
		delete(partitionConsumers, p)
	}

	return nil
}

func (h demoTdsqlConsumerGroupHandler) Cleanup(_ sarama.ConsumerGroupSession) error {
	fmt.Println("cleanup")
	return nil
}

func (h demoTdsqlConsumerGroupHandler) ConsumeClaim(sess sarama.ConsumerGroupSession, claim sarama.ConsumerGroupClaim) error {
	for msg := range claim.Messages() {
		if err := partitionConsumers[msg.Partition].consume(sess, msg); nil != err {
			return err
		}
	}
	return nil
}

type tdsqlPartitionMsgConsumer struct {
	completeMsg     *subscribe.Envelope
	shardLastSeqMap map[string]uint64
	shardMsgMap     map[string]*subscribe.Envelope
	verbose         int
}

func (tc *tdsqlPartitionMsgConsumer) consume(sess sarama.ConsumerGroupSession, msg *sarama.ConsumerMessage) error {
	// Each message has a sequence id in the partition, this id is continuous monotonic increasing,
	// these messages are duplicated if the id rotate;
	// There is unexpected message lost in kafka if the id is discontinuous,
	// have to reset the subscribe and client when it happened.
	// When the task is restarted, the sequence id will start from 1
	ps := getPartitionSeq(msg)

	// Each message has a shard id,
	// this id can be used to distinguish that the message belongs to which shard
	shardId := getShardId(msg)
	shardLastSeq, ok := tc.shardLastSeqMap[shardId]
	if !ok || ps == uint64(1) {
		tc.shardLastSeqMap[shardId] = uint64(1)
	} else if ps <= shardLastSeq {
		fmt.Printf("duplicated msg: %+v\n", msg)
	} else if ps != shardLastSeq+1 && shardLastSeq != 0 {
		msg := fmt.Sprintf("the partition seq discontinuous, last: %v, current: %v", shardLastSeq, ps)
		fmt.Printf("%s\n", msg)
		panic(msg) //kafka write lost
	}

	tc.shardLastSeqMap[shardId] = ps
	envelope := &subscribe.Envelope{}
	err := proto.Unmarshal(msg.Value, envelope)
	if nil != err {
		fmt.Println(err)
		return err
	}

	if tc.verbose == PRINT_KAFKAMETA {
		fmt.Println("kafka msg===>", "Partition:", msg.Partition, "Offset:", msg.Offset, "time:", msg.Timestamp, "ps:", ps, "Index:", envelope.Index, "Total:", envelope.Total, "Properties:", envelope.Properties, "data_len:", len(envelope.Data))
	}

	if int32(1) != envelope.Version {
		fmt.Printf("unsupported version: %v, message: %+v\n", envelope.Version, envelope)
		return errors.New("unsupported version")
	}

	// The size of one event maybe exceed the size limitation of kafka message,
	// then the event will be split into more than one message,
	// each message has an index, the value starts from zero to total-1,
	// have to rebuild the byte array before decode the object by protobuf.
	if envelope.Total == 1 {
		tc.completeMsg = envelope
	} else {
		shardMsg, ok := tc.shardMsgMap[shardId]
		if !ok {
			tc.shardMsgMap[shardId] = envelope
		} else {
			shardMsg.Data = append(shardMsg.Data, envelope.Data...)
		}
	}

	// index starts from zero
	if envelope.Index < envelope.Total-1 {
		return nil
	}

	if tc.verbose == PRINT_KAFKAMETA {
		fmt.Println("envelope received total===>", "Index:", envelope.Index, "Total:", envelope.Total, "Properties:", envelope.Properties, "data_len:", len(envelope.Data))
	}

	entries := subscribe.Entries{}
	if envelope.Total == 1 {
		err = proto.Unmarshal(tc.completeMsg.Data, &entries)
	} else {
		err = proto.Unmarshal(tc.shardMsgMap[shardId].Data, &entries)
		// Pay attention to memory recycling for shardMsgMap
		delete(tc.shardMsgMap, shardId)
	}
	if nil != err {
		fmt.Println(err)
		return err
	}

	if tc.verbose == PRINT_DMLOPT {
		for _, entry := range entries.Items {
			if err = onEntry(msg.Partition, msg.Offset, ps, entry); nil != err {
				fmt.Println(err)
				return err
			}
		}
	}

	// commit at checkpoint message
	if len(entries.Items) > 0 && entries.Items[0].Header.MessageType == subscribe.MessageType_CHECKPOINT {
		sess.MarkMessage(msg, "") //自动和手动：先标记再提交; 手动提交模式下，也需要先进行标记
	}

	return nil
}

const partitionSeqKey = "ps"

func getPartitionSeq(msg *sarama.ConsumerMessage) uint64 {
	for _, m := range msg.Headers {
		if string(m.Key) == partitionSeqKey {
			n, err := strconv.ParseUint(string(m.Value), 10, 64)
			if nil != err {
				panic("unexpected msg seq: " + string(m.Value))
			}

			return n
		}
	}

	panic(partitionSeqKey + " does not exists")
}

const shardKey = "ShardId"

func getShardId(msg *sarama.ConsumerMessage) string {
	for _, m := range msg.Headers {
		if string(m.Key) == shardKey {
			fmt.Printf("shardId[%s]\n", string(m.Value))
			return string(m.Value)
		}
	}
	panic(shardKey + " does not exists")
}

func onEntry(partitionNum int32, offset int64, partitionSeq uint64, entry *subscribe.Entry) error {
	if *trans2sql {
		return trans2SQL(partitionNum, offset, partitionSeq, entry)
	} else {
		fmt.Printf("partition: %v, offset: %v, partitionSeq: %v, %+v\n", partitionNum, offset, partitionSeq, entry)
		return nil
	}
}

func quoteIdentifiers(identifier string) string {
	strb := strings.Builder{}
	if strings.Contains(identifier, "\"") {
		replace := strings.Replace(identifier, "\\", "\\005c", -1)
		replace = strings.Replace(replace, "\"", "\\0022", -1)
		strb.WriteString("U&\"")
		strb.WriteString(replace)
		strb.WriteString("\"")
	} else {
		strb.WriteString("\"")
		strb.WriteString(identifier)
		strb.WriteString("\"")
	}

	return strb.String()
}

func trans2SQL(partitionNum int32, offset int64, partitionSeq uint64, entry *subscribe.Entry) error {
	header := entry.Header
	fmt.Printf("-->[partition: %d, offset: %d, partitionSeq: %d] [%s:%d], happenedAt: %s\n",
		partitionNum, offset, partitionSeq, header.FileName, header.Position, time.Unix(int64(header.Timestamp), 0))

	switch entry.Header.MessageType {
	case subscribe.MessageType_HEARTBEAT:
		fmt.Println("heartbeat:", entry.Event.HeartbeatEvent.Epoch)
	case subscribe.MessageType_CHECKPOINT:
		evt := entry.Event.CheckpointEvent
		fmt.Println("Checkpoint: gtid=", evt.SyncedGtid, ", file=", evt.FileName, ", position=", evt.Position)
	case subscribe.MessageType_BEGIN:
		fmt.Println("BEGIN")
	case subscribe.MessageType_COMMIT:
		fmt.Println("COMMIT")
	case subscribe.MessageType_ROLLBACK:
		fmt.Println("ROLLBACK")
	case subscribe.MessageType_DDL:
		fmt.Println(entry.Event.DdlEvent.Sql)
	case subscribe.MessageType_DML:
		dmlEvt := entry.Event.DmlEvent
		switch dmlEvt.DmlEventType {
		case subscribe.DMLType_INSERT:
			if err := transInsert(entry); nil != err {
				return err
			}
		case subscribe.DMLType_DELETE:
			if err := transDelete(entry); nil != err {
				return err
			}
		case subscribe.DMLType_UPDATE:
			if err := transUpdate(entry); nil != err {
				return err
			}
		default:
			return errors.New("unsupported dml event type")
		}
	default:
		return errors.New("unsupported message type")
	}

	return nil
}

func transInsert(entry *subscribe.Entry) error {
	dmlEvt := entry.Event.DmlEvent
	fmt.Printf("INSERT INTO %v.%v VALUES ", quoteIdentifiers(entry.Header.SchemaName), quoteIdentifiers(entry.Header.TableName))
	rowSep := ""
	for _, row := range dmlEvt.Rows {
		fmt.Print(rowSep)
		fmt.Print("(")
		sep := ""
		for _, col := range row.NewColumns {
			fmt.Print(sep)
			if d, err := decode(col); nil != err {
				fmt.Println(err)
				return err
			} else {
				fmt.Printf("%v", d)
			}
			sep = ", "
		}

		fmt.Print(")")
		rowSep = ", "
	}

	fmt.Println("")
	return nil
}

func transDelete(entry *subscribe.Entry) error {
	dmlEvt := entry.Event.DmlEvent
	header := fmt.Sprintf("DELETE FROM %v.%v WHERE ", quoteIdentifiers(entry.Header.SchemaName), quoteIdentifiers(entry.Header.TableName))
	for _, row := range dmlEvt.Rows {
		fmt.Print(header)
		sep := ""
		for idx, col := range row.OldColumns {
			if col.DataType == subscribe.DataType_NA {
				continue
			}

			fmt.Print(sep)
			colDef := dmlEvt.Columns[idx]
			fmt.Printf("%v", quoteIdentifiers(colDef.Name))
			if col.DataType == subscribe.DataType_NIL {
				fmt.Print(" IS NULL")
			} else if d, err := decode(col); nil != err {
				fmt.Println(err)
				return err
			} else {
				fmt.Printf(" = %v", d)
			}
			sep = " AND "
		}

		fmt.Println("")
	}

	return nil
}

func transUpdate(entry *subscribe.Entry) error {
	dmlEvt := entry.Event.DmlEvent
	header := fmt.Sprintf("UPDATE %v.%v SET ", quoteIdentifiers(entry.Header.SchemaName), quoteIdentifiers(entry.Header.TableName))
	values := make([]string, 0)
	conditions := make([]string, 0)
	for _, row := range dmlEvt.Rows {
		for idx, col := range row.NewColumns {
			if col.DataType == subscribe.DataType_NA {
				continue
			}

			colDef := dmlEvt.Columns[idx]
			if !colDef.IsKey {
				if d, err := decode(col); nil != err {
					fmt.Println(err)
					return err
				} else {
					values = append(values, fmt.Sprintf("%v = %v", quoteIdentifiers(colDef.Name), d))
				}
			}
		}

		for _, col := range row.OldColumns {
			if col.DataType == subscribe.DataType_NA {
				continue
			}
			colName := quoteIdentifiers(col.Name)
			if col.DataType == subscribe.DataType_NIL {
				conditions = append(conditions, fmt.Sprintf("%v IS NULL", colName))
			} else if d, err := decode(col); nil != err {
				fmt.Println(err)
				return err
			} else {
				conditions = append(conditions, fmt.Sprintf("%v = %v", colName, d))
			}
		}

		sql := header + strings.Join(values, ", ") + " WHERE " + strings.Join(conditions, " AND ")
		fmt.Println(sql)
	}

	return nil
}

func decode(data *subscribe.Data) (string, error) {
	if numberTypes.Contains(data.DataType) {
		return data.Sv, nil
	}
	switch data.DataType {
	case subscribe.DataType_STRING:
		return "'" + string(data.Bv) + "'", nil
	case subscribe.DataType_BYTES:
		return "'\\x" + hex.EncodeToString(data.Bv) + "'", nil
	case subscribe.DataType_NA:
		return "DEFAULT", nil
	case subscribe.DataType_NIL:
		return "NULL", nil
	default:
		return "", errors.New("unsupported data type")
	}
}
